# Copyright (c) 2016 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
import logging
import hashlib
import json

from pymongo.errors import DuplicateKeyError


logger = logging.getLogger(__name__)

KEY_CHECKSUM = 'checksum'
KEY_ELEMENT = 'element'


def encode_element(element):
    binary = json.dumps(json.loads(
        json.dumps(element)), sort_keys=True).encode()
    return binary


def decode_element(binary):
    return json.loads(binary.decode())


def calculate_checksum(binary):
    m = hashlib.sha1()
    m.update(binary)
    return m.hexdigest()


class MongoDBSet(object):

    def __init__(self, collection):
        """Initialize an instance of MongoDBSet.

        :param collection: The MongoDB collection to be used for this set.
        """
        self._collection = collection

    def __len__(self):
        """Return the size of the set."""
        return self._collection.find().count()

    def _filter(self, elem):
        return {'_id': calculate_checksum(encode_element(elem))}

    def __contains__(self, elem):
        """Returns true if elem is a member of the set."""
        return self._collection.find_one(self._filter(elem)) is not None

    def add(self, elem):
        """Add element to the set, if not already present."""
        binary = encode_element(elem)
        try:
            decoded = decode_element(binary)
            if elem != decoded:
                raise ValueError((elem, decoded))
        except Exception as error:
            msg = "Attempt to encode element '{}' failed: {}."
            raise ValueError(msg.format(elem, error))
        checksum = calculate_checksum(binary)
        doc = {
            '_id': checksum,
            KEY_ELEMENT: binary}
        try:
            self._collection.insert_one(doc)
        except DuplicateKeyError:
            pass

    def remove(self, elem):
        """Remove element from the set.

        :raises KeyError if elem is not member of the set."""
        result = self._collection.delete_one(self._filter(elem))
        if not result.deleted_count:
            raise KeyError(elem)

    def discard(self, elem):
        """Remove element from the set if it is present."""
        self._collection.delete_one(self._filter(elem))

    def pop(self):
        """Remove and return an arbitrary element from the set.

        :raises KeyError if the set is empty.
        """
        document = self._collection.find_one_and_delete({})
        if document is None:
            raise KeyError()
        else:
            return decode_element(document[KEY_ELEMENT])

    def clear(self):
        """Remove all elements from the set."""
        self._collection.delete_many({})

    def __iter__(self):
        for doc in self._collection.find():
            yield decode_element(doc[KEY_ELEMENT])



# Copyright (c) 2016 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
import warnings
from .moab import MoabScheduler
from .slurm import SlurmScheduler
from .fakescheduler import FakeScheduler


try:
    from .apscheduler import APScheduler  # noqa
except ImportError:
    warnings.warn("Failed to import apscheduler. "
                  "The test scheduler will not be available.",
                  ImportWarning)

    class APScheduler(object):
        """This is a mock class.

        Install apscheduler to enable the test environment."""

        def __init__(self, *args, **kwargs):
            raise ImportError("Install apscheduler to enable this scheduler.")

__all__ = ['MoabScheduler', 'SlurmScheduler', 'APScheduler', 'FakeScheduler']



# Copyright (c) 2016 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
"""Define environments to automate submission scripts.

Partially adapted from clusterutils package by
Matthew Spellings."""

from __future__ import print_function
import re
import socket
import logging
from signac.common.six import with_metaclass


logger = logging.getLogger(__name__)

MODE_CPU = 'cpu'
MODE_GPU = 'gpu'


class ComputeEnvironmentType(type):

    def __init__(cls, name, bases, dct):
        if not hasattr(cls, 'registry'):
            cls.registry = dict()
        else:
            cls.registry[name] = cls
        return super(ComputeEnvironmentType, cls).__init__(name, bases, dct)


class ComputeEnvironment(with_metaclass(ComputeEnvironmentType)):
    hostname_pattern = None

    def __init__(self, mode=MODE_CPU):
        if mode not in self.header_scripts.keys():
            raise ValueError(mode)
        self.mode = mode

    @classmethod
    def is_present(cls):
        if cls.hostname_pattern is None:
            return False
        else:
            return re.match(
                cls.hostname_pattern, socket.gethostname()) is not None

    @classmethod
    def submit(cls, jobsid, np, walltime, script,
               pretend=False, *args, **kwargs):
        raise NotImplementedError()


class UnknownEnvironment(ComputeEnvironment):
    pass


class TestEnvironment(ComputeEnvironment):
    pass


class MoabEnvironment(ComputeEnvironment):
    submit_cmd = ['qsub']
    cores_per_node = None


class SlurmEnvironment(ComputeEnvironment):
    cores_per_node = None


class CPUEnvironment(ComputeEnvironment):
    pass


class GPUEnvironment(ComputeEnvironment):
    pass


def get_environment():
    for env_type in ComputeEnvironment.registry.values():
        if env_type.is_present():
            return env_type
    else:
        return UnknownEnvironment



# Copyright (c) 2016 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
"""Workflow management based on the signac framework.

The signac-flow package provides the basic infrastructure
to easily configure workflow for data operations on a
signac workspace executed by a scheduler."""
from . import environment
from . import scheduler
from . import manage
from . project import FlowProject

__version__ = '0.1.9'

__all__ = [
    'environment',
    'scheduler',
    'manage',
    'FlowProject'
]



__all__ = ['tqdm', 'trange']

import sys
import time


def format_interval(t):
    mins, s = divmod(int(t), 60)
    h, m = divmod(mins, 60)
    if h:
        return '%d:%02d:%02d' % (h, m, s)
    else:
        return '%02d:%02d' % (m, s)


def format_meter(n, total, elapsed):
    # n - number of finished iterations
    # total - total number of iterations, or None
    # elapsed - number of seconds passed since start
    if n > total:
        total = None

    elapsed_str = format_interval(elapsed)
    rate = '%5.2f' % (n / elapsed) if elapsed else '?'

    if total:
        frac = float(n) / total

        N_BARS = 10
        bar_length = int(frac*N_BARS)
        bar = '#'*bar_length + '-'*(N_BARS-bar_length)

        percentage = '%3d%%' % (frac * 100)

        left_str = format_interval(elapsed / n * (total-n)) if n else '?'

        return '|%s| %d/%d %s [elapsed: %s left: %s, %s iters/sec]' % (
            bar, n, total, percentage, elapsed_str, left_str, rate)

    else:
        return '%d [elapsed: %s, %s iters/sec]' % (n, elapsed_str, rate)


class StatusPrinter(object):
    def __init__(self, file):
        self.file = file
        self.last_printed_len = 0

    def print_status(self, s):
        self.file.write('\r'+s+' '*max(self.last_printed_len-len(s), 0))
        self.file.flush()
        self.last_printed_len = len(s)


def tqdm(iterable, desc='', total=None, leave=False, file=sys.stderr,
         mininterval=0.5, miniters=1):
    """
    Get an iterable object, and return an iterator which acts exactly like the
    iterable, but prints a progress meter and updates it every time a value is
    requested.
    'desc' can contain a short string, describing the progress, that is added
    in the beginning of the line.
    'total' can give the number of expected iterations. If not given,
    len(iterable) is used if it is defined.
    'file' can be a file-like object to output the progress message to.
    If leave is False, tqdm deletes its traces from screen after it has
    finished iterating over all elements.
    If less than mininterval seconds or miniters iterations have passed since
    the last progress meter update, it is not updated again.
    """
    if total is None:
        try:
            total = len(iterable)
        except TypeError:
            total = None

    prefix = desc+': ' if desc else ''

    sp = StatusPrinter(file)
    sp.print_status(prefix + format_meter(0, total, 0))

    start_t = last_print_t = time.time()
    last_print_n = 0
    n = 0
    for obj in iterable:
        yield obj
        # Now the object was created and processed, so we can print the meter.
        n += 1
        if n - last_print_n >= miniters:
            # We check the counter first, to reduce the overhead of time.time()
            cur_t = time.time()
            if cur_t - last_print_t >= mininterval:
                sp.print_status(prefix + format_meter(n, total, cur_t-start_t))
                last_print_n = n
                last_print_t = cur_t

    if not leave:
        sp.print_status('')
        sys.stdout.write('\r')
    else:
        if last_print_n < n:
            cur_t = time.time()
            sp.print_status(prefix + format_meter(n, total, cur_t-start_t))
        file.write('\n')


def trange(*args, **kwargs):
    """A shortcut for writing tqdm(range()) on py3 or tqdm(xrange()) on py2"""
    try:
        f = xrange
    except NameError:
        f = range

    return tqdm(f(*args), **kwargs)



# Copyright (c) 2016 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
"""Routines for the MOAB environment."""

from __future__ import print_function
import io
import getpass
import subprocess
import tempfile
import math
import logging
import xml.etree.ElementTree as ET

from .manage import Scheduler
from .manage import ClusterJob, JobStatus


logger = logging.getLogger(__name__)


def _fetch(user=None):
    if user is None:
        user = getpass.getuser()
    cmd = "qstat -fx -u {user}".format(user=user)
    try:
        result = io.BytesIO(subprocess.check_output(cmd.split()))
    except FileNotFoundError:
        raise RuntimeError("Moab not available.")
    tree = ET.parse(source=result)
    return tree.getroot()


def format_timedelta(delta):
    hours, r = divmod(delta.seconds, 3600)
    minutes, seconds = divmod(r, 60)
    hours += delta.days * 24
    return "{:0>2}:{:0>2}:{:0>2}".format(hours, minutes, seconds)


class MoabJob(ClusterJob):

    def __init__(self, node):
        self.node = node

    def _id(self):
        return self.node.find('Job_Id').text

    def __str__(self):
        return str(self._id())

    def name(self):
        return self.node.find('Job_Name').text

    def status(self):
        job_state = self.node.find('job_state').text
        if job_state == 'R':
            return JobStatus.active
        if job_state == 'Q':
            return JobStatus.queued
        if job_state == 'C':
            return JobStatus.inactive
        if job_state == 'H':
            return JobStatus.held
        return JobStatus.registered


class MoabScheduler(Scheduler):
    submit_cmd = ['qsub']

    def __init__(self, root=None, user=None, header=None, cores_per_node=None):
        if root is None:
            root = _fetch(user=user)
        self.root = root
        self.header = header
        self.cores_per_node = cores_per_node

    def jobs(self):
        for node in self.root.findall('Job'):
            yield MoabJob(node)

    def submit(self, jobsid, np, walltime, script, resume=None,
               after=None, pretend=False, hold=False, *args, **kwargs):
        submit_script = io.StringIO()
        num_nodes = math.ceil(np / self.cores_per_node)
        if (np / (num_nodes * self.cores_per_node)) < 0.9:
            logger.warning("Bad node utilization!")
        submit_script.write(self.header.format(
            jobsid=jobsid, nn=num_nodes, walltime=format_timedelta(walltime)))
        submit_script.write('\n')
        submit_script.write(script.read())
        submit_script.seek(0)
        submit = submit_script.read().format(
            np=np, nn=num_nodes,
            walltime=format_timedelta(walltime), jobsid=jobsid)
        if pretend:
            print("#\n# Pretend to submit:\n")
            print(submit, "\n")
        else:
            submit_cmd = self.submit_cmd
            if after is not None:
                submit_cmd.extend(
                    ['-W', 'depend="afterok:{}"'.format(after.split('.')[0])])
            if hold:
                submit_cmd += ['-h']
            with tempfile.NamedTemporaryFile() as tmp_submit_script:
                tmp_submit_script.write(submit.encode('utf-8'))
                tmp_submit_script.flush()
                output = subprocess.check_output(
                    submit_cmd + [tmp_submit_script.name])
            jobsid = output.decode('utf-8').strip()
            return jobsid



# Copyright (c) 2016 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
import logging
import time
import queue
import itertools
from threading import Thread, Event
from math import tanh
from itertools import count

from pymongo import ASCENDING


logger = logging.getLogger(__name__)

ID_COUNTER = 0
KEY_COUNTER = 'counter'
FILTER_COUNTER = {'_id': ID_COUNTER}
FILTER_COUNTER_DEC = {'_id': ID_COUNTER, KEY_COUNTER: {'$gt': 0}}
FILTER_NOT_COUNTER = {'_id': {'$ne': ID_COUNTER}}


def _fetch(target, timeout=None, stop_event=None):
    tmp_queue = queue.Queue()
    if stop_event is None:
        stop_event = Event()

    def inner_loop():
        w = (tanh(0.05 * i) for i in itertools.count())
        while(not stop_event.is_set()):
            result = target()
            if result is not None:
                tmp_queue.put(result)
                return
        stop_event.wait(max(0.001, next(w)))
    thread__fetch = Thread(target=inner_loop)
    thread__fetch.start()
    try:
        thread__fetch.join(timeout=timeout)
    except KeyboardInterrupt:
        stop_event.set()
        thread__fetch.join()
        raise
    if thread__fetch.is_alive():
        stop_event.set()
        thread__fetch.join()
    try:
        return tmp_queue.get_nowait()
    except queue.Empty:
        raise TimeoutError()


class Empty(Exception):
    pass


class Full(Exception):
    pass


class MongoDBQueue(object):

    def __init__(self, collection):
        self._collection = collection

    def qsize(self):
        return self._collection.find(FILTER_NOT_COUNTER).count()

    def empty(self):
        return self.qsize() == 0

    def full(self):
        return False  # There is no limit on the queue size implemented.

    def _get(self):
        return self._collection.find_one_and_delete(
            FILTER_NOT_COUNTER, sort=[('_id', ASCENDING)])

    def __contains__(self, item):
        return self._collection.find_one(item) is not None

    def _num_open_tasks(self):
        result = self._collection.find_one(FILTER_COUNTER)
        if result is None:
            return 0
        else:
            open_tasks = int(result[KEY_COUNTER])
            if open_tasks < 0:
                raise ValueError()
            else:
                return open_tasks

    def put(self, item, block=True, timeout=None):
        # block and timeout are ignored, as in this implementation, the queue
        # can never be full.
        result = self._collection.insert_one(item)
        self._collection.update_one(
            FILTER_COUNTER, {'$inc': {KEY_COUNTER: 1}}, upsert=True)
        return result.inserted_id

    def get(self, block=True, timeout=None, stop_event=None):
        if block:
            try:
                return _fetch(
                    target=self._get,
                    timeout=timeout,
                    stop_event=stop_event)
            except TimeoutError:
                raise Empty()
        else:
            item = self._get()
            if item is None:
                raise Empty()
            else:
                return item

    def get_nowait(self):
        return self.get(block=False)

    def task_done(self):
        result = self._collection.update_one(
            FILTER_COUNTER_DEC, {'$inc': {KEY_COUNTER: -1}})
        if result.modified_count != 1:
            raise ValueError()

    def join(self):
        w = (tanh(0.05 * i) for i in count())
        while True:
            if self._num_open_tasks() == 0:
                return
            time.sleep(max(0.001, next(w)))

    def peek(self):
        for item in self._collection.find(FILTER_NOT_COUNTER):
            yield item

    def clear(self):
        self._collection.delete_many(
            {'$or': [FILTER_COUNTER, FILTER_NOT_COUNTER]})



# Copyright (c) 2016 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
from __future__ import print_function
import logging

from .manage import Scheduler

logger = logging.getLogger(__name__)


class FakeScheduler(Scheduler):

    def jobs(self):
        return
        yield

    def submit(self, jobsid, np, walltime, script, *args, **kwargs):
        logger.info("Fake scheduling of job '{}' on {} procs with "
                    "a walltime of '{}'.".format(jobsid, np, walltime))
        for line in script:
            print(line)
        return jobsid



#!/usr/bin/env python3
# Copyright (c) 2016 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.

import logging
import enum


logger = logging.getLogger(__name__)


class JobStatus(enum.IntEnum):
    """Classifies the job's execution status.

    The stati are ordered by the significance
    of the execution status.
    This enables easy comparison, such as

    .. code-block: python

        if status < JobStatus.submitted:
            submit()

    which prevents a submission of a job,
    which is already submitted, queued, active
    or in an error state."""
    unknown = 1
    registered = 2
    inactive = 3
    submitted = 4
    held = 5
    queued = 6
    active = 7
    error = 8
# User stati are >= 128.
    user = 128


class ClusterJob(object):

    def __init__(self, jobid, status=None):
        self._job_id = jobid
        self._status = status

    def _id(self):
        return self._job_id

    def __str__(self):
        return str(self._id())

    def name(self):
        return self._id()

    def status(self):
        return self._status


class Scheduler(object):
    "Generic Scheduler ABC"

    def jobs(self):
        "yields ClusterJob"
        raise NotImplementedError()


def _status_local(jobsid):
    """Attempt to determine status with local information."""
    return JobStatus.unknown


def _status_scheduler(jobsid, scheduler_jobs):
    """Attempt to determine status with information from the scheduler."""
    cjob = scheduler_jobs.get(jobsid)
    if cjob is None:
        status = JobStatus.unknown
    else:
        status = max(JobStatus.registered, cjob.status())
    return status


def update_status(job, scheduler_jobs=None):
    """Update the job's status dictionary."""
    status_doc = job.document.setdefault('status', dict())
    for jobsid in status_doc.keys():
        status = _status_local(jobsid)
        if scheduler_jobs is not None:
            status = max(_status_scheduler(jobsid, scheduler_jobs), status)
        status_doc[jobsid] = int(status)
        job.document['status'] = status_doc


def submit(env, project, state_point, script,
           identifier='default', force=False, pretend=False,
           *args, **kwargs):
    """Attempt to submit a job to the scheduler of the current environment.

    The job status will be determined from the job's status document.
    If the job's status is greater or equal than JobStatus.submitted,
    the job will not be submitted, unless the force option is provided."""
    job = project.open_job(state_point)
    # A jobsid is a unique identifier used to identify this job with
    # the job scheduler.
    jobsid = '{}-{}-{}'.format(project, job, identifier)
    logger.info("Attempting submittal of job '{}'.".format(job))
    logger.debug("Determine status...")

    def set_status(value):
        "Update the job's status dictionary."
        status_doc = job.document.get('status', dict())
        status_doc[jobsid] = int(value)
        job.document['status'] = status_doc
    try:
        status = job.document['status'][jobsid]
    except KeyError:
        set_status(JobStatus.registered)
        status = job.document['status'][jobsid]
    if not force:
        if status >= JobStatus.submitted:
            logger.info(
                "Job currently blocked from submission "
                "(already submitted or active).")
            return False
    try:
        assert pretend or env.submit(
            jobsid=jobsid, script=script, pretend=pretend, *args, **kwargs)
    except Exception:
        logger.warning("Error.")
        set_status(JobStatus.error)
        raise
    else:
        set_status(JobStatus.submitted)
        logger.info("Success.")
        return True



# Copyright (c) 2016 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
from . import tabulate, tqdm

__all__ = ['tabulate', 'tqdm']



#!/usr/bin/env python3
# Copyright (c) 2016 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
"""Routines for the APScheduler environment."""

from __future__ import print_function
import logging
import subprocess
import tempfile
import argparse
import random
import time
from datetime import datetime, timedelta

import signac
import apscheduler
from apscheduler.schedulers.blocking import BlockingScheduler
from apscheduler.jobstores.mongodb import MongoDBJobStore
from apscheduler.executors.pool import ProcessPoolExecutor, ThreadPoolExecutor
from apscheduler.triggers.date import DateTrigger

from .mongodb_queue import MongoDBQueue, Empty
from .mongodb_set import MongoDBSet
from .manage import Scheduler, ClusterJob, JobStatus


logger = logging.getLogger(__name__)


def format_timedelta(delta):
    hours, r = divmod(delta.seconds, 3600)
    minutes, seconds = divmod(r, 60)
    hours += delta.days * 24
    return "{:0>2}:{:0>2}:{:0>2}".format(hours, minutes, seconds)


def get_queue(db, name):
    project = signac.get_project()
    q_name = "{project}_scheduler_{name}".format(
        project=project, name=name)
    return MongoDBQueue(db[q_name])
    # return MongoDBQueue(signac.db.get_database(db)[q_name])


def get_set(name, db=None):
    project = signac.get_project()
    if db is None:
        db = str(project)
    q_name = "{project}-scheduler-{name}".format(
        project=project, name=name)
    return MongoDBSet(signac.db.get_database(db)[q_name])


def transfer(item, src, dst):
    get_set(dst).add(item)
    get_set(src).discard(item)


def execute_script(jobsid, np, walltime, script, *args, **kwargs):
    with tempfile.NamedTemporaryFile() as tmp:
        tmp.write(script.encode())
        tmp.flush()
        with open('{}.log'.format(jobsid), 'ab') as logfile:
            try:
                cmd = '/bin/sh {}'.format(tmp.name)
                logger.debug("Executing command '{}'".format(cmd))
                subprocess.check_call(
                    cmd,
                    stdout=logfile,
                    stderr=logfile,
                    shell=True)
            except subprocess.CalledProcessError as error:
                logger.warning(error)
            finally:
                logfile.flush()


def run_job(scheduler, job_doc):
    try:
        transfer(job_doc['jobsid'], 'queued', 'active')
        execute_script(** job_doc)
    except Exception as error:
        logger.error(error)
        raise
    finally:
        transfer(job_doc['jobsid'], 'active', 'completed')


def submit_job(scheduler, queue, misfire_grace_time, next_job):
    run_date = next_job.get('run_date', datetime.now())
    run_date += timedelta(seconds=10+int(random.random() * 10))
    after = next_job.get('after')
    assert after is None
    trigger = DateTrigger(run_date=run_date)
    scheduler.add_job(
        run_job,
        id=next_job['jobsid'],
        args=(scheduler, next_job),
        #executor='processpool',
        trigger=trigger,
        misfire_grace_time=misfire_grace_time)
    return next_job['jobsid']


def poll_submits(scheduler, queue, misfire_grace_time=None):
    logger.debug("Pulling from submittal queue.")
    # Allow delay of job start up to an hour later to prevent skipping
    if misfire_grace_time is None:
        misfire_grace_time = 3600
    while True:
        try:
            next_job = queue.get(block=False)
            if next_job.get('after') is None:
                submit_job(scheduler, queue, misfire_grace_time, next_job)
            else:
                jobid = 'submitafter-{}'.format(next_job['after'])
                scheduler.add_job(
                    submit_after,
                    id=jobid,
                    args=(jobid, scheduler, queue, misfire_grace_time, next_job),
                    executor='threadpool',
                    trigger='interval',
                    seconds=20)
            get_set('queued').add(next_job['jobsid'])
        except Empty:
            break
        else:
            queue.task_done()
            time.sleep(1)


def submit_after(jobid, scheduler, queue, misfire_grace_time, next_job):
    after = next_job['after']
    if after in get_set('completed'):
        submit_job(scheduler, queue, misfire_grace_time, next_job)
        scheduler.get_job(jobid).remove()


def show_jobs(scheduler):
    print(scheduler.get_jobs())


class APSJob(ClusterJob):
    pass


class APScheduler(Scheduler):

    def __init__(self, db):
        self.scheduler = get_scheduler()
        self.db = db

    def _get_submit_queue(self):
        return get_queue(self.db, 'submitted')

    def jobs(self):
        stati = dict()
        for job in self._get_submit_queue().peek():
            stati[job['jobsid']] = JobStatus.submitted
        for name, status in (
                ('queued', JobStatus.queued),
                ('active', JobStatus.active),
                ('completed', JobStatus.inactive)):
            for job_id in get_set(name):
                stati[job_id] = max(
                    stati.get(job_id, JobStatus.registered), status)
                yield APSJob(job_id, status)
        for job_id, status in stati.items():
            yield APSJob(job_id, status)

    def submit(self, jobsid, np, walltime, script,
               pretend=False, after=None, **kwargs):
        job_doc = {
            'jobsid': jobsid,
            'np': np,
            'walltime': walltime.seconds,
            'after': after,
            'script': script.read().format(
                jobsid=jobsid, np=np, walltime=format_timedelta(walltime)),
        }
        if pretend:
            print(type(self), "Pretend to submit:\n\n", job_doc)
            return jobsid
        else:
            logger.info("Submitting '{}'.".format(job_doc))
            self._get_submit_queue().put(job_doc)
            return jobsid


def get_scheduler(num_procs=1):
    return BlockingScheduler(
        jobstores={
            'mongo': MongoDBJobStore(
                client=signac.common.host.get_client(),
                database=str(signac.get_project()),
                collection='apscheduler')},
        executors={
            'threadpool': ThreadPoolExecutor(),
            #'processpool': ProcessPoolExecutor(max_workers=num_procs)
            },
    )


def clear_sets():
    logger.debug("Clearing internal state.")
    for name in ('queued', 'active', 'completed'):
        get_set(name).clear()


def reset(db):
    logger.info("Resetting all queues.")
    get_queue(db, 'submitted').clear()
    clear_sets()


def start_scheduler(num_procs, db):
    scheduler = get_scheduler(num_procs)
    clear_sets()
    print("Starting scheduler.")
    scheduler.add_job(
        poll_submits,
        args=(scheduler, get_queue(db, 'submitted')),
        executor='threadpool',
        trigger='cron',
        second='*/10')
    # scheduler.add_job(
    #     show_jobs,
    #     args=(scheduler,),
    #     trigger='cron',
    #     second='*/12')
    try:
        scheduler.start()
    except KeyboardInterrupt:
        print("Shutting down scheduler. ", end='')
        print("Hit ctrl-c again to exit immediately.")
        scheduler.shutdown(wait=True)
        print("Scheduler shut down.")
    except BaseException:
        scheduler.shutdown(wait=False)
        raise


def main(args):
    db = signac.db.get_database(args.db)
    if args.reset:
        reset(db)
    else:
        start_scheduler(num_procs=args.num_procs, db=db)


if __name__ == '__main__':
    parser = argparse.ArgumentParser(
        description="Execute a test scheduler. "
                    "Do not start more than one instance!")
    parser.add_argument(
        '-n', '--num-procs',
        type=int,
        default=1,
        help="The number of processors to use for execution.")
    parser.add_argument(
        '--reset',
        action='store_true',
        help="Do not start the scheduler, but reset all internal queues.")
    parser.add_argument(
        '--db',
        type=str,
        default=str(signac.get_project()),
        help="Specify which database to use.")
    args = parser.parse_args()
    logging.basicConfig(level=logging.INFO)
    main(args)



# Copyright (c) 2016 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
"""Routines for the MOAB environment."""

from __future__ import print_function
import io
import getpass
import subprocess
import tempfile
import math
import logging

from .manage import Scheduler
from .manage import ClusterJob, JobStatus


logger = logging.getLogger(__name__)


def _fetch(user=None):
    def parse_status(s):
        if s == 'Q':
            return JobStatus.queued
        elif s == 'R':
            return JobStatus.active
        elif s == 'C':
            return JobStatus.inactive
        else:
            assert 0

    if user is None:
        user = getpass.getuser()
    cmd = "qstat -f"
    try:
        result = subprocess.check_output(cmd.split()).decode()
    except subprocess.CalledProcessError as error:
        if error.returncode == 153:
            return
        else:
            raise
    except FileNotFoundError:
        raise RuntimeError("Slurm not available.")
    jobs = result.split('\n\n')
    for jobinfo in jobs:
        if not jobinfo:
            continue
        lines = jobinfo.split('\n')
        assert lines[0].startswith('Job Id:')
        info = dict()
        for line in lines[1:]:
            tokens = line.split('=')
            info[tokens[0].strip()] = tokens[1].strip()
        if info['euser'].startswith(user):
            yield SlurmJob(info['Job_Name'], parse_status(info['job_state']))


def format_timedelta(delta):
    hours, r = divmod(delta.seconds, 3600)
    minutes, seconds = divmod(r, 60)
    hours += delta.days * 24
    return "{:0>2}:{:0>2}:{:0>2}".format(hours, minutes, seconds)


class SlurmJob(ClusterJob):
    pass


class SlurmScheduler(Scheduler):
    submit_cmd = ['sbatch']

    def __init__(self, user=None, header=None, cores_per_node=None):
        self.header = header
        self.cores_per_node = cores_per_node
        self.user = user

    def jobs(self):
        for job in _fetch(user=self.user):
            yield job

    def submit(self, jobsid, np, walltime, script, resume=None,
               after=None, pretend=False, *args, **kwargs):
        submit_script = io.StringIO()
        num_nodes = math.ceil(np / self.cores_per_node)
        if (np / (num_nodes * self.cores_per_node)) < 0.9:
            logger.warning("Bad node utilization!")
        submit_script.write(self.header.format(
            jobsid=jobsid, nn=num_nodes, walltime=format_timedelta(walltime)))
        submit_script.write('\n')
        submit_script.write(script.read())
        submit_script.seek(0)
        submit = submit_script.read().format(
            np=np, nn=num_nodes, walltime=format_timedelta(walltime), jobsid=jobsid)
        if pretend:
            print("#\n# Pretend to submit:\n")
            print(submit, "\n")
        else:
            submit_cmd = self.submit_cmd
            if after is not None:
                submit_cmd.extend(
                    ['-W', 'depend="afterany:{}"'.format(after.split('.')[0])])
            with tempfile.NamedTemporaryFile() as tmp_submit_script:
                tmp_submit_script.write(submit.encode('utf-8'))
                tmp_submit_script.flush()
                output = subprocess.check_output(
                    submit_cmd + [tmp_submit_script.name])
            return jobsid



# Copyright (c) 2016 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
import sys

if sys.version_info < (2, 7, 0):
    print("Error: signac-flow requires python version >= 2.7.x.")
    sys.exit(1)

requirements = []
if sys.version_info < (3, 4, 0):
    requirements.append('enum34')

from setuptools import setup, find_packages

setup(
    name='signac-flow',
    version='0.2.0',
    packages=find_packages(),
    zip_safe=True,

    author='Carl Simon Adorf',
    author_email='csadorf@umich.edu',
    description="Simple workflow management based on signac.",
    keywords='workflow management singac framework database',

    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Topic :: Scientific/Engineering :: Physics",
    ],

    install_requires=requirements,
)



# Copyright (c) 2016 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
from __future__ import print_function
import sys
import os
import io
import logging
import datetime
import json
from collections import defaultdict
from itertools import islice
from hashlib import sha1

import signac

from . import manage
from . import util
from .util.tqdm import tqdm


logger = logging.getLogger(__name__)

DEFAULT_WALLTIME_HRS = 12


def is_active(status):
    for gid, s in status.items():
        if s > manage.JobStatus.inactive:
            return True
    return False


def draw_progressbar(value, total, width=40):
    n = int(value / total * width)
    return '|' + ''.join(['#'] * n) + ''.join(['-'] * (width - n)) + '|'

def _update_status(args):
    return manage.update_status(* args)


class JobScript(io.StringIO):
    "Simple StringIO wrapper to implement cmd wrapping logic."

    def writeline(self, line, eol='\n'):
        "Write one line to the job script."
        self.write(line + eol)

    def write_cmd(self, cmd, parallel=False, np=1, mpi_cmd=None, **kwargs):
        """Write a command to the jobscript.

        This command wrapper function is a convenience function, which
        adds mpi and other directives whenever necessary.

        The ``mpi_cmd`` argument should be a callable, with the following
        signature: ``mpi_cmd(cmd, np, **kwargs)``.

        :param cmd: The command to write to the jobscript.
        :type cmd: str
        :param parallel: Commands should be executed in parallel.
        :type parallel: bool
        :param np: The number of processors required for execution.
        :type np: int
        :param mpi_cmd: MPI command wrapper.
        :type mpi_cmd: callable
        :param kwargs: All other forwarded parameters."""
        if np > 1:
            if mpi_cmd is None:
                raise RuntimeError("Requires mpi_cmd wrapper.")
            cmd = mpi_cmd(cmd, np=np)
        if parallel:
            cmd += ' &'
        self.writeline(cmd)
        return np


class FlowProject(signac.contrib.Project):
    """A signac project class assisting in job scheduling.

    :param config: A signac configuaration, defaults to
        the configuration loaded from the environment.
    :type config: A signac config object."""

    def _get_jobsid(self, job, operation):
        "Return a unique job session id based on the job and operation."
        return '{jobid}-{operation}'.format(jobid=job, operation=operation)

    def _store_bundled(self, job_ids):
        """Store all job session ids part of one bundle.

        The job session ids are stored in a text file in the project's
        root directory. This is necessary to be able to identify each
        job's individual status from the bundle id."""
        sid = '{self}-bundle-{sid}'.format(
            self=self,
            sid=sha1('.'.join(job_ids).encode('utf-8')).hexdigest())
        with open(os.path.join(self.root_directory(), sid), 'w') as file:
            for job_id in job_ids:
                file.write(job_id + '\n')
        return sid

    def _expand_bundled_jobs(self, scheduler_jobs):
        "Expand jobs which were submitted as part of a bundle."
        for job in scheduler_jobs:
            if job.name().startswith('{}-bundle-'.format(self)):
                fn_bundle = os.path.join(self.root_directory(), job.name())
                with open(fn_bundle) as file:
                    for line in file:
                        yield manage.ClusterJob(line.strip(), job.status())
            else:
                yield job

    def _fetch_scheduler_jobs(self, scheduler):
        """Fetch jobs from the scheduler.

        This function does not select specific jobs, but relies on the
        scheduler implementation to make a pre-selection of jobs, which
        might be associated with this project."""
        return {job.name(): job
                for job in self._expand_bundled_jobs(scheduler.jobs())}

    def _update_status(self, job, scheduler_jobs):
        "Determine the scheduler status of job."
        manage.update_status(job, scheduler_jobs)

    def get_job_status(self, job):
        "Return the detailed status of jobs."
        result = dict()
        result['job_id'] = str(job)
        status = job.document.get('status', dict())
        result['active'] = is_active(status)
        result['labels'] = sorted(set(self.classify(job)))
        result['operation'] = self.next_operation(job)
        highest_status = max(status.values()) if len(status) else 1
        result['submission_status'] = [manage.JobStatus(highest_status).name]
        return result

    def _blocked(self, job, operation, **kwargs):
        "Check if job, operation combination is blocked for scheduling."
        try:
            status = job.document['status'][self._get_jobsid(job, operation)]
            return status >= manage.JobStatus.submitted
        except KeyError:
            return False

    def _eligible(self, job, operation=None, **kwargs):
        """Internal check for the job's eligible for operation.

        A job is only eligible if the public :meth:`~.eligible` method
        returns True and the job is not blocked by the scheduler.

        :raises RuntimeError: If the public eligible method returns None."""
        ret = self.eligible(job, operation, **kwargs) \
            and not self._blocked(job, operation, **kwargs)
        if ret is None:
            raise RuntimeError("Unable to determine eligiblity for job '{}' "
                               "and job type '{}'.".format(job, operation))
        return ret

    def _submit(self, scheduler, to_submit, pretend,
                serial, bundle, after, walltime, **kwargs):
        "Submit jobs to the scheduler."
        script = JobScript()
        self.write_header(
            script=script, walltime=walltime, serial=serial,
            bundle=bundle, after=after, ** kwargs)
        jobids_bundled = []
        np_total = 0
        for job, operation in to_submit:
            jobsid = self._get_jobsid(job, operation)

            def set_status(value):
                "Update the job's status dictionary."
                status_doc = job.document.get('status', dict())
                status_doc[jobsid] = int(value)
                job.document['status'] = status_doc
                return int(value)

            np = self.write_user(
                script=script, job=job, operation=operation,
                parallel=not serial and bundle is not None, **kwargs)
            if np is None:
                raise RuntimeError(
                    "Failed to return 'num_procs' value in write_user()!")
            np_total = max(np, np_total) if serial else np_total + np
            if pretend:
                set_status(manage.JobStatus.registered)
            else:
                set_status(manage.JobStatus.submitted)
            jobids_bundled.append(jobsid)
        script.write('wait')
        script.seek(0)
        if not len(jobids_bundled):
            return False

        if len(jobids_bundled) > 1:
            sid = self._store_bundled(jobids_bundled)
        else:
            sid = jobsid
        scheduler_job_id = scheduler.submit(
            script=script, jobsid=sid,
            np=np_total, walltime=walltime, pretend=pretend, **kwargs)
        logger.info("Submitted {}.".format(sid))
        if serial and not bundle:
            if after is None:
                after = ''
            after = ':'.join(after.split(':') + [scheduler_job_id])
        return True

    def to_submit(self, job_ids=None, operation=None, job_filter=None):
        """Generate a sequence of (job_id, operation) value pairs for submission.

        :param job_ids: A list of job_id's,
            defaults to all jobs found in the workspace.
        :param operation: A specific operation,
            defaults to the result of :meth:`~.next_operation`.
        :param job_filter: A JSON encoded filter,
            that all jobs to be submitted need to match."""
        if job_ids is None:
            jobs = list(self.find_jobs(job_filter))
        else:
            jobs = [self.open_job(id=jobid) for jobid in job_ids]
        if operation is None:
            operations = (self.next_operation(job) for job in jobs)
        else:
            operations = [operation] * len(jobs)
        return zip(jobs, operations)

    def filter_non_eligible(self, to_submit, **kwargs):
        "Return only those jobs for submittal, which are eligible."
        return ((j, jt) for j, jt in to_submit
                if self._eligible(j, jt, **kwargs))

    def submit_jobs(self, scheduler, to_submit, walltime=None,
                    bundle=None, serial=False, after=None,
                    num=None, pretend=False, force=False, **kwargs):
        """Submit jobs to the scheduler.

        :param scheduler: The scheduler instance.
        :type scheduler: :class:`~.flow.manage.Scheduler`
        :param to_submit: A sequence of (job_id, operation) tuples.
        :param walltime: The maximum wallclock time in hours.
        :type walltime: float
        :param bundle: Bundle up to 'bundle' number of jobs during submission.
        :type bundle: int
        :param serial: Schedule jobs in serial or execute bundled
            jobs in serial.
        :type serial: bool
        :param after: Execute all jobs after the completion of the job operation
            with this job session id. Implementation is scheduler dependent.
        :type after: str
        :param num: Do not submit more than 'num' jobs to the scheduler.
        :type num: int
        :param pretend: Do not actually submit, but instruct the scheduler
            to pretend scheduling.
        :type pretend: bool
        :param force: Ignore all eligibility checks, just submit.
        :type force: bool
        :param kwargs: Other keyword arguments which are forwareded."""
        if walltime is not None:
            walltime = datetime.timedelta(hours=walltime)
        if not force:
            to_submit = self.filter_non_eligible(to_submit, **kwargs)
        to_submit = islice(to_submit, num)
        if bundle is not None:
            n = None if bundle == 0 else bundle
            while True:
                ts = islice(to_submit, n)
                if not self._submit(scheduler, ts, walltime=walltime,
                                    bundle=bundle, serial=serial, after=after,
                                    num=num, pretend=pretend, force=force, **kwargs):
                    break
        else:
            for ts in to_submit:
                self._submit(scheduler, [ts], walltime=walltime,
                             bundle=bundle, serial=serial, after=after,
                             num=num, pretend=pretend, force=force, **kwargs)

    def submit(self, scheduler, job_ids=None,
               operation=None, job_filter=None, **kwargs):
        """Wrapper for :meth:`~.to_submit` and :meth:`~.submit_jobs`.

        This function passes the return value of :meth:`~.to_submit`
        to :meth:`~.submit_jobs`.

        :param scheduler: The scheduler instance.
        :type scheduler: :class:`~.flow.manage.Scheduler`
        :param job_ids: A list of job_id's,
            defaults to all jobs found in the workspace.
        :param operation: A specific operation,
            defaults to the result of :meth:`~.next_operation`.
        :param job_filter: A JSON encoded filter that all jobs
            to be submitted need to match.
        :param kwargs: All other keyword arguments are forwarded
            to :meth:`~.submit_jobs`."""
        if job_ids is not None and not len(job_ids):
            job_ids = None
        return self.submit_jobs(
            scheduler=scheduler,
            to_submit=self.to_submit(job_ids, operation, job_filter),
            **kwargs)

    @classmethod
    def add_submit_args(cls, parser):
        "Add arguments to parser for the :meth:`~.submit` method."
        parser.add_argument(
            'job_ids',
            type=str,
            nargs='*',
            help="The job id of the jobs to submit. "
            "Omit to automatically select all eligible jobs.")
        parser.add_argument(
            '-j', '--job-operation',
            dest='operation',
            type=str,
            help="Only submit jobs eligible for the specified operation.")
        parser.add_argument(
            '-w', '--walltime',
            type=float,
            default=DEFAULT_WALLTIME_HRS,
            help="The wallclock time in hours.")
        parser.add_argument(
            '--pretend',
            action='store_true',
            help="Do not really submit, but print the submittal script.")
        parser.add_argument(
            '-n', '--num',
            type=int,
            help="Limit the number of jobs submitted at once.")
        parser.add_argument(
            '--force',
            action='store_true',
            help="Do not check job status or classification, just submit.")
        parser.add_argument(
            '-f', '--filter',
            dest='job_filter',
            type=str,
            help="Filter jobs.")
        parser.add_argument(
            '--after',
            type=str,
            help="Schedule this job to be executed after "
                 "completion of job with this id.")
        parser.add_argument(
            '-s', '--serial',
            action='store_true',
            help="Schedule the jobs to be executed serially.")
        parser.add_argument(
            '--bundle',
            type=int,
            nargs='?',
            const=0,
            help="Specify how many jobs to bundle into one submission. "
                 "Omit a specific value to bundle all eligible jobs.")
        parser.add_argument(
            '--hold',
            action='store_true',
            help="Submit job with user hold applied.")

    def write_human_readable_statepoint(self, script, job):
        "Write statepoint of job in human-readable format to script."
        script.write('# Statepoint:\n#\n')
        sp_dump = json.dumps(job.statepoint(), indent=2).replace(
            '{', '{{').replace('}', '}}')
        for line in sp_dump.splitlines():
            script.write('# ' + line + '\n')
        script.write('\n')

    def write_header(self, script, walltime, env=None, **kwargs):
        """Write a general jobscript header to the script.

        The header is written only once for each job script, whereas the
        output of :meth:`~.write_user` may be written multiple times to
        one job script for bundled jobs.

        :param script: The job script, to write to.
        :type script: :class:`~.JobScript`
        :param walltime: The maximum allowed walltime for this operation.
        :type walltime: :class:`datetime.timedelta`

        The default method writes nothing."""
        return

    def write_user(self, script, job, operation, parallel, **kwargs):
        """Write to the jobscript for job and job type."

        This function needs to be specialized for each project.
        Please note, that all commands should obey the parallel flag, i.e.
        should be executed in parallel if set to True.
        You should `assert not parallel` if that is not possible.

        See also: :meth:`~.JobScript.write_cmd`.

        The ``mpi_cmd`` argument should be a callable, with the following
        signature: ``mpi_cmd(cmd, np, **kwargs)``.

        :param script: The job script, to write to.
        :type script: :class:`~.JobScript`
        :param job: The signac job handle.
        :type job: :class:`signac.contrib.Job`
        :param operation: The name of the operation to execute.
        :type operation: str
        :param parallel: Execute commands in parallel if True.
        :type parallel: bool
        :param mpi_cmd: MPI command wrapper. Pass this function
            to execute mpi commands in multiple environments.

        :returns: The number of required processors (nodes).
        :rtype: int
        """
        raise NotImplementedError()

    def print_overview(self, stati, file=sys.stdout):
        "Print the project's status overview."
        progress = defaultdict(int)
        for status in stati:
            for label in status['labels']:
                progress[label] += 1
        progress_sorted = sorted(
            progress.items(), key=lambda x: (x[1], x[0]), reverse=True)
        table_header = ['label', 'progress']
        rows = ([label, '{} {:0.2f}%'.format(
            draw_progressbar(num, len(stati)), 100 * num / len(stati))]
            for label, num in progress_sorted)
        print("Total # of jobs: {}".format(len(stati)), file=file)
        print(util.tabulate.tabulate(rows, headers=table_header), file=file)

    def format_row(self, status, statepoint=None):
        "Format each row in the detailed status output."
        row = [
            status['job_id'],
            ', '.join(status['submission_status']),
            status['operation'],
            ', '.join(status.get('labels', [])),
        ]
        if statepoint:
            sps = self.open_job(id=status['job_id']).statepoint()
            for i, sp in enumerate(statepoint):
                row.insert(i + 1, sps.get(sp))
        return row

    def print_detailed(self, stati, parameters=None,
                       skip_active=False, file=sys.stdout):
        "Print the project's detailed status."
        table_header = ['job_id', 'status', 'next_operation', 'labels']
        if parameters:
            for i, value in enumerate(parameters):
                table_header.insert(i + 1, value)
        rows = (self.format_row(status, parameters)
                for status in stati if not (skip_active and status['active']))
        print(util.tabulate.tabulate(rows, headers=table_header), file=file)

    def export_job_stati(self, collection, stati):
        for status in stati:
            job = self.open_job(id=status['job_id'])
            status['statepoint'] = job.statepoint()
            collection.update_one({'_id': status['job_id']},
                                  {'$set': status}, upsert=True)

    def update_stati(self, scheduler, jobs=None, file=sys.stderr, pool=None):
        """Update the status of all jobs with the given scheduler.

        :param scheduler: The scheduler instance used to feth the job stati.
        :type scheduler: :class:`~.manage.Scheduler`
        :param jobs: A sequence of :class:`~.signac.contrib.Job` instances.
        :param file: The file to write output to, defaults to `sys.stderr`."""
        if jobs is None:
            jobs = self.find_jobs()
        print("Query scheduler...", file=file)
        scheduler_jobs = self._fetch_scheduler_jobs(scheduler)
        print("Determine job stati...", file=file)
        if pool is None:
            for job in tqdm(jobs, file=file):
                self._update_status(job, scheduler_jobs)
        else:
            jobs_ = ((job, scheduler_jobs) for job in jobs)
            pool.map(_update_status, tqdm(jobs_, total=len(jobs), file=file))

    def print_status(self, scheduler=None, job_filter=None,
                     detailed=False, parameters=None, skip_active=False,
                     file=sys.stdout, err=sys.stderr,
                     pool=None):
        """Print the status of the project.

        :param scheduler: The scheduler instance used to fetch the job stati.
        :type scheduler: :class:`~.manage.Scheduler`
        :param job_filter: A JSON encoded filter,
            that all jobs to be submitted need to match.
        :param detailed: Print a detailed status of each job.
        :type detailed: bool
        :param parameters: Print the value of the specified parameters.
        :type parameters: list of str
        :param skip_active: Only print jobs that are currently inactive.
        :type skip_active: bool
        :param file: Print all output to this file,
            defaults to sys.stdout
        :param err: Print all error output to this file,
            defaults to sys.stderr
        :param pool: A multiprocessing or threading pool. Providing a pool
            parallelizes this method."""
        if job_filter is not None and isinstance(job_filter, str):
            job_filter = json.loads(job_filter)
        jobs = list(self.find_jobs(job_filter))
        if scheduler is not None:
            self.update_stati(scheduler, jobs, file=err, pool=pool)
        print("Generate output...", file=err)
        if pool is None:
            stati = [self.get_job_status(job) for job in jobs]
        else:
            stati = pool.map(self.get_job_status, jobs)
        title = "Status project '{}':".format(self)
        print('\n' + title, file=file)
        self.print_overview(stati)
        if detailed:
            print(file=file)
            print("Detailed view:", file=file)
            self.print_detailed(stati, parameters, skip_active, file)

    @classmethod
    def add_print_status_args(cls, parser):
        "Add arguments to parser for the :meth:`~.print_status` method."
        parser.add_argument(
            '-f', '--filter',
            dest='job_filter',
            type=str,
            help="Filter jobs.")
        parser.add_argument(
            '-d', '--detailed',
            action='store_true',
            help="Display a detailed view of the job stati.")
        parser.add_argument(
            '-p', '--parameters',
            type=str,
            nargs='*',
            help="Display select parameters of the job's "
                 "statepoint with the detailed view.")
        parser.add_argument(
            '--skip-active',
            action='store_true',
            help="Display only jobs, which are currently not active.")

    def classify(self, job):
        """Generator function which yields labels for job.

        :param job: The signac job handle.
        :type job: :class:`~signac.contrib.job.Job`
        :yields: The labels to classify job.
        :yield type: str"""

    def next_operation(self, job):
        """Determine the next operation for job.

        You can, but don't have to use this function to simplify
        the submission process. The default method returns None.

        :param job: The signac job handle.
        :type job: :class:`~signac.contrib.job.Job`
        :returns: The name of the operation to execute next.
        :rtype: str"""
        return

    def eligible(self, job, operation=None, **kwargs):
        """Determine if job is eligible for operation."""
        return None
